/*
    Credits:
    Thank you to Julie Wang and Ellen Chuang for their code and documentation 
    for the HC-05 Bluetooth module.

    Switches to include in Linker Options:
	-Wl,-u,vfprintf -lprintf_flt -lm
	-Wl,-u,vfscanf -lscanf_flt -lm
 */

#define F_CPU 16000000UL
#include <util/delay.h>
#include <avr/sleep.h>
#include <stdio.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h>
#include <float.h>
#include <limits.h>
#include <stdlib.h>
#include "trtSettings.h"
#include "trtkernel_1284.c"

//Bit macros
#define set_bit(address,bit) (address |= (1<<bit))
#define clear_bit(address,bit) (address &= ~(1<<bit))
#define toggle_bit(address,bit) (address ^= (1<<bit))
#define check_bit(address,bit) ((address & (1<<bit)) == (1<<bit))

// serial communication library
// Don't mess with the semaphores
#define SEM_RX_ISR_SIGNAL 1
#define SEM_STRING_DONE 2 // user hit <enter>
#include "trtUart.h"
#include "trtUart.c"

// two semaphores to protect message --
// sender must wait until message is received 
// -- init to 1 because don't need to wait for first message
// receiver must wait until message is sent
#define SEM_TX_WAIT  4
#define SEM_RX_WAIT  5

// semaphore to protect shared variable
#define SEM_SHARED_VAR 7

#define CCW 0    //counterclockwise direction for motor
#define CW 1     //clockwise direction for motor
#define leftMotor 0
#define rightMotor 1

#define RIGHT_FORWARD CW
#define RIGHT_REVERSE CCW
#define LEFT_FORWARD CCW
#define LEFT_REVERSE CW

//Set to 0 if using UART to communicate over Bluetooth; 1 if printing to PuTTY
#define DEBUG 0

// UART file descriptor
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);
int args[1]; //not actually used, just a dummy for trtCreateTask

volatile int x_angle;
volatile int y_angle;
volatile uint8_t leftMotorPWM = 20;
volatile uint8_t rightMotorPWM = 20;

//whether to emergency stop the car if Bluetooth connection is lost
//the only way to clear it is to reset the car
volatile char stopCar = 0;


//========== Forward declarations ==========
void initialize(void);      //sets up interrupts
void motorDrive(uint8_t, uint8_t, uint8_t);
void motorDriveTask(void* args);
void bluetoothTask(void* args);

//*********** External Interrupt ISR ***************
ISR (INT1_vect) {
	
	//triggered when D2 changes
	if (check_bit(PORTD, 2)) {
		//Bluetooth module became connected
		stopCar = 0;
		fprintf(stdout, "int connect\n");
	}
	else {
		//Bluetooth module become DISCONNECTED
		stopCar = 1;
		fprintf(stdout, "int disconnect\n");
	}
}
    
void initialize(void){

	//init the UART
	trt_uart_init();
	stdout = stdin = stderr = &uart_str;
    fprintf(stdout,"\n\r UART initialized\n\r\n\r");

	////PORT A
	//| | | |BIN1|AIN1|STBY|BIN2|AIN2|
	DDRA = 0b00011111;
	PORTA = 0;

	//Set B0-B6 as outputs
    DDRB = 0b00011000;
	PORTB = 0;
	
	//set D3 to input
	clear_bit(DDRD, 3);

	//Set up Timer 0
    //divide by 64- 16000000Hz/64/256 bits = 976 Hz
    TCCR0B = 0b011; 

	// turn on fast PWM and OC0A output at full clock rate,
    // toggle OC0A (pin B.3) and OC0B (pin B.4)
    // 16 microsec per PWM cycle sample time
	// set COM0A0 to 0, COM0A1 to 1 - noninverting mode
    TCCR0A = (1<<COM0A1) | (1<<COM0B1) | (1<<WGM00) | (1<<WGM01) ; 
    OCR0A = 128;
	OCR0B = 128;

	//set up INT0 to check when Bluetooth module becomes disconnected
    //EIMSK = 1 << INT1 ; // turn on int0
    //EICRA = 1 << 1SC10; // generate interrupt on either edge, pg 327, 65

	sei();  // Interrupts - leggo!
}

void motorDrive(uint8_t motorNumber, uint8_t motorDirection, uint8_t motorSpeed){	
	
	//Check if you want motor stopped
	if (motorSpeed == 0) {
		if (motorNumber == leftMotor){
			PORTA &= !((1<<PA0)|(1<<PA3)); //Set AIN2 & AIN1 to low
			PORTA |= (1<<PA2);	//Set STBY to high
			PORTB |= (1<<PB3); //Set PWMA to high
		}
		else {
			PORTA &= !((1<<PA1)|(1<<PA4)); //Set BIN2 & BIN1 to low
			PORTA |= (1<<PA2);	//Set STBY to high
			PORTB |= (1<<PB4); //Set PWMB to high			
		}
	}
	else{
		//Set direction for left motor
		if (motorNumber == leftMotor){
			if (motorDirection == CCW){
				//must clear first because AIN1 & 2 can't both be 1, bad for h-bridge
				clear_bit(PORTA, PA0);		//AIN2 set to 0
				PORTA |= (1<<PA3)|(1<<PA2); //AIN1 set to 1
			}
			else {
				clear_bit(PORTA, PA3);		//AIN1 set to 0
				PORTA |= (1<<PA0)|(1<<PA2); //AIN2 set to 1
			}
			OCR0A = motorSpeed;
		}
		else { //Set direction for  right motor
			if (motorDirection == CCW){
				clear_bit(PORTA, PA1);		//BIN2 set to 0
				PORTA |= (1<<PA4)|(1<<PA2); //BIN1 set to 1
			}
			else{
				clear_bit(PORTA, PA4);		//BIN1 set to 0
				PORTA |= (1<<PA1)|(1<<PA2); //BIN2 set to 1
			}
			OCR0B = motorSpeed;
		}
	} //end of stop/go motor if

 } //end of motorDrive()

void motorDriveTask(void* args){
	uint32_t rel, dead;

	char leftDir  = CCW;
	char rightDir = CW;

	int8_t leftTmp  = 0;
	int8_t rightTmp = 0;

	#define MAX_ANGLE 75

	while (1) {
		if (!check_bit(PIND, 3)) {
			//set PWM values to 0
			motorDrive(leftMotor,  CCW,  0);
			motorDrive(rightMotor, CW, 0);
			fprintf(stdout, "d3 is low\n");
		}

		else {
		//fprintf(stdout, "this is working\n");

		//===== x_angle controls forward/backward =====
		//+x = reverse,    -x = forward
		
		//trtWait(SEM_SHARED_VAR);
		if (-10 < x_angle && x_angle < 10) {
			//take within 10 degrees of horizontal to be 0
			leftMotorPWM = 0;
			rightMotorPWM = 0;
		}
		else if (x_angle <= -10) {
			//move forward
			leftMotorPWM = ((double)-x_angle)/MAX_ANGLE*255;
			rightMotorPWM = ((double)-x_angle)/MAX_ANGLE*255;
			leftDir  = LEFT_FORWARD;
			rightDir = RIGHT_FORWARD;
		}
		else if (x_angle >= 10) {
			//move backwards
			leftMotorPWM = ((double)x_angle)/MAX_ANGLE*255;
			rightMotorPWM = ((double)x_angle)/MAX_ANGLE*255;
			leftDir  = LEFT_REVERSE;
			rightDir = RIGHT_REVERSE;
		}


		//===== y_angle controls left/right =====
		//+y = right turn, -y = left turn
		
		if (-10 < y_angle && y_angle < 10) {
			//setting both wheels to same speed should make the car not turn
			leftMotorPWM = rightMotorPWM;
		}
		else if (y_angle >= 10){
			//turning right
			leftDir  = (x_angle < 0) ? LEFT_FORWARD : LEFT_REVERSE;

			rightTmp = (int8_t)leftMotorPWM - (int8_t)(((float)y_angle)/MAX_ANGLE*2*leftMotorPWM);
			rightDir = (rightTmp > 0) ? RIGHT_FORWARD : RIGHT_REVERSE; //have R wheel go backwards if difference is big enough
			if (x_angle > 0) rightDir = -rightDir; //reverse direction if car is going backwards

			rightMotorPWM = (rightTmp > 0) ? rightTmp : -rightTmp; //PWM value is absolute value
		}
		else if (y_angle <= -10) {
			//turning left
			leftTmp = (int8_t)rightMotorPWM - (int8_t)(((float)-y_angle)/MAX_ANGLE*2*rightMotorPWM);
			leftDir  = (leftTmp > 0) ? LEFT_FORWARD : LEFT_REVERSE; //have L wheel go backwards if difference is big enough
			if (x_angle > 0) leftDir = -leftDir; //reverse direction if car is going backwards

			leftMotorPWM = (leftTmp > 0) ? leftTmp : -leftTmp; //PWM value is absolute value

			rightDir = (x_angle < 0) ? RIGHT_FORWARD : RIGHT_REVERSE;
		}
		//trtSignal(SEM_SHARED_VAR);
		
		//bounds checking
		if (leftMotorPWM > 255) leftMotorPWM = 255;
		if (rightMotorPWM > 255) rightMotorPWM = 255;

		//move in the specified direction
		motorDrive(leftMotor,  leftDir,  leftMotorPWM);
		motorDrive(rightMotor, rightDir, rightMotorPWM);

		} //end else

		// Sleep
	    rel = trtCurrentTime() + SECONDS2TICKS(0.15);
	    dead = trtCurrentTime() + SECONDS2TICKS(0.25);
		trtSleepUntil(rel, dead);
	}
}

void bluetoothTask(void* args){

	char lcd_buffer[17];

	while(1){
		fscanf(stdin, "%s", lcd_buffer);
		
		//trtWait(SEM_SHARED_VAR);
		sscanf(lcd_buffer,"%d,%d", &x_angle, &y_angle);
		//fprintf(stdout,"(%d, %d)\n", x_angle, y_angle);
		//trtSignal(SEM_SHARED_VAR);
	}
}
	
int main(void)
{
	initialize();
	
	// start TRT
  	trtInitKernel(80); // 80 bytes for the idle task stack

	// --- create semaphores ----------
	// You must creat the first two semaphores if you use the uart
	trtCreateSemaphore(SEM_RX_ISR_SIGNAL, 0) ; // uart receive ISR semaphore
	trtCreateSemaphore(SEM_STRING_DONE,0) ;  // user typed <enter>

	// message protection
	trtCreateSemaphore(SEM_TX_WAIT, 1) ; // message send interlock
	trtCreateSemaphore(SEM_RX_WAIT, 0) ; // message receive interlock

	// variable protection
	trtCreateSemaphore(SEM_SHARED_VAR, 1) ; // protect shared variable between putty & motor control

	// --- create tasks  ----------------
	trtCreateTask(motorDriveTask, 250, 
		SECONDS2TICKS(0.05), SECONDS2TICKS(0.05), &(args[0]));
	trtCreateTask(bluetoothTask, 500, 
		SECONDS2TICKS(0.1), SECONDS2TICKS(0.3), &(args[0]));


	// --- Idle task --------------------------------------
	// just sleeps the cpu to save power every time it executes
	set_sleep_mode(SLEEP_MODE_IDLE);
	sleep_enable();

    while(1)
    {
		sleep_cpu();
		
		/*
		//Forward
		motorDrive(leftMotor, CCW, leftMotorPWM);
		motorDrive(rightMotor,CW, rightMotorPWM);

		//Reverse
		motorDrive(leftMotor, CW, leftMotorPWM);
		motorDrive(rightMotor,CCW, rightMotorPWM);

		//Left
		motorDrive(leftMotor, CW, leftMotorPWM);
		motorDrive(rightMotor,CW, rightMotorPWM);

		//Right
		motorDrive(leftMotor, CCW, leftMotorPWM);
		motorDrive(rightMotor,CCW, rightMotorPWM);
		*/
    }
}
